//=========================================================================//
//-------------------------------------------------------------------------//
// $Id: i2cbus.cpp,v 1.3 2001/12/12 05:43:54 lanconel Exp $                                                                    //
//-------------------------------------------------------------------------//
//  Copyright (C) 1998-2001  Claudio Lanconelli                            //
//                                                                         //
//  e-mail: lancos@libero.it                                               //
//  http://www.LancOS.com                                                  //
//-------------------------------------------------------------------------//
//                                                                         //
// This program is free software; you can redistribute it and/or           //
// modify it under the terms of the GNU  General Public License            //
// as published by the Free Software Foundation; either version2 of        //
// the License, or (at your option) any later version.                     //
//                                                                         //
// This program is distributed in the hope that it will be useful,         //
// but WITHOUT ANY WARRANTY; without even the implied warranty of          //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       //
// General Public License for more details.                                //
//                                                                         //
// You should have received a copy of the GNU  General Public License      //
// along with this program (see COPYING);     if not, write to the         //
// Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. //
//                                                                         //
//-------------------------------------------------------------------------//
//=========================================================================//

#include "types.h"
#include "errcode.h"
#include "i2cbus.h"

#ifdef	_LINUX_
#  include <unistd.h>
#else
#  ifdef	__BORLANDC__
#    define	__inline__
#  else // _MICROSOFT_ VC++
#    define	__inline__ __inline
#    define _export
#  endif
#endif

#undef	SCLTIMEOUT	// disabilita il controllo e la temporizzazione su SCL

#define BUSYDELAY	100
#define SDATIMEOUT	200

// Costruttore
I2CBus::I2CBus()
{
	errno = last_addr = 0;
	shot_delay = 5;
}

// Distruttore
I2CBus::~I2CBus()
{
	Close();
}

int I2CBus::CheckBusy()
{
	register int count;

	for (count = BUSYDELAY; count > 0 && isSCLSDAuno(); count--)
		WaitUsec(1);
	return (count ? IICERR_BUSBUSY : 0);
}

int I2CBus::SendStart()
{
	register int err;

	setSCLSDA();
#ifdef	SCLTIMEOUT
	if (getSCL() == 0)
	{
		int k;

		for(k = SCLTIMEOUT; getSCL() == 0 && k > 0; k--)
			WaitUsec(1);
		if (k == 0)
			return IICERR_SCLCONFLICT;
	}
#else
 	WaitUsec(shot_delay/2);
#endif
	if ( (err = CheckBusy()) )
		return err;
	clearSDA();
	WaitUsec(shot_delay);	// tHD;STA = 4 usec
	clearSCL();
	WaitUsec(shot_delay/2);

	return 0;
}

int I2CBus::SendStop()
{
	clearSCL();
	WaitUsec(shot_delay+1);
	clearSDA();
	WaitUsec(shot_delay+1);
	setSCL();
#ifdef	SCLTIMEOUT
	if (getSCL() == 0)
	{
		int k;

		for(k = SCLTIMEOUT; getSCL() == 0 && k > 0; k--)
			WaitUsec(1);
		if (k == 0)
		{
		//	UserDebug(UserApp2, "I2CBus::SendStop() *** SCL error\n");
			return IICERR_SCLCONFLICT;
		}
	}
#else
	WaitUsec(shot_delay/2);
#endif
	WaitUsec(shot_delay+1);	// tSU;STOP = 4.7 usec
	setSDA();
	WaitUsec(shot_delay/2+1);
	if (getSDA() == 0)
	{
	//	UserDebug(UserApp2, "I2CBus::SendStop() *** SDA error\n");
		return IICERR_SDACONFLICT;
	}

	//tBUF = 4.7 usec
	WaitUsec(shot_delay);

	return 0;
}

int I2CBus::SendBitMast(int b)
{
	bitSDA(b);
	WaitUsec(shot_delay/2+1);	// tSU;DAT = 250 nsec (tLOW / 2 = 2 usec)
	setSCL();
	/* Se SCL e` ancora 0 significa che uno Slave sta` rallentando
	 * il trasferimento, attendi che SCL si rialzi */
	if (getSCL() == 0)
	{
#ifdef	SCLTIMEOUT
		int k;

		for(k = SCLTIMEOUT; getSCL() == 0 && k > 0; k--)
			WaitUsec(1);
		if (k == 0)
			return IICERR_SCLCONFLICT;
#endif
	}
	WaitUsec(shot_delay/2);	// tHIGH / 2 = 2 usec
	if (!getSDA() != !b)
		return IICERR_SDACONFLICT;
	WaitUsec(shot_delay/2);	// tHIGH / 2 = 2 usec
	clearSCL();
	WaitUsec(shot_delay/2);	// tHD;DATA = 300 nsec (tLOW / 2 = 2 usec)

	return 0;
}

// returns a negative number if there's an error, 0 or 1 otherwise
int I2CBus::RecBitMast()
{
	register UBYTE b;

	setSDA();		// to receive data SDA must be high
	WaitUsec(shot_delay/2+1);	// tSU;DAT = 250 nsec (tLOW / 2 = 2 usec)
	setSCL();
	/* Se SCL e` ancora 0 significa che uno Slave sta` rallentando
	 * il trasferimento, attendi che SCL si rialzi */
	if (getSCL() == 0)
	{
#ifdef	SCLTIMEOUT
		int k;

		for(k = SCLTIMEOUT; getSCL() == 0 && k > 0; k--)
			WaitUsec(1);
		if (k == 0)
			return IICERR_SCLCONFLICT;
#endif
	}
	WaitUsec(shot_delay/2);	// tHIGH / 2 = 2 usec
	b = getSDA();
	WaitUsec(shot_delay/2);	// tHIGH / 2 = 2 usec
	clearSCL();
	WaitUsec(shot_delay/2);	// tHD;DATA = 300 nsec (tLOW / 2 = 2 usec)

	return b;
}

// OK, ora ci alziamo di un livello: operiamo sul byte
int I2CBus::SendByteMast(int by)
{
	int lrb,k;

	for (k = 7; k >= 0; k--)
		if ( (lrb = SendBitMast(by & (1<<k))) )
			return lrb;

	lrb = RecBitMast();	// acknowledge bit
	if (lrb < 0)		// < 0 means that an error occured
		return lrb;
	if (lrb)
		return IICERR_NOTACK;

	return 0;
}

int I2CBus::RecByteMast(int ack)
{
	int k,lrb,val = 0;

	for (k = 7; k >= 0; k--)
	{
		lrb = RecBitMast();
		if (lrb < 0)
			return lrb;
		if (lrb)
			val |= 1 << k;
	}
	if ( (k = SendBitMast(ack)) )	/* manda l'ack */
		return k;
	setSDA();	/* rilascia SDA allo slave trasmitter */

	return val;
}

ULONG I2CBus::Read(int slave, UBYTE *data, ULONG length)
{
	ULONG len;

	len = StartRead(slave, data, length);
	if (len == length)
		if ( errno || Stop() )
			len = 0;
	return len;
}

ULONG I2CBus::Write(int slave, UBYTE const *data, ULONG length)
{
	ULONG len;

	len = StartWrite(slave, data, length);
	if (len == length)
		if ( errno || Stop())
			len = 0;
	return len;
}

int I2CBus::ReadByte(int ack)
{
	return RecByteMast(ack);
}


int I2CBus::Start(UBYTE slave)
{
	int temp;

	/* Invio Start */
	if ( (temp = SendStart()) )
	{
		errno = temp;
		return errno;
	}

	if ( (temp = SendByteMast(slave)) != 0 )
	{
		errno = (temp == IICERR_NOTACK) ? IICERR_NOADDRACK : temp;
		last_addr = slave;
		return errno;
	}

	return 0;
}

/* se length e` zero, viene ricevuto un byte senza ack per cercare di
 * liberare il bus da uno slave che per qualche motivo e` ancora in
 * trasmissione. In questo caso data deve puntare ad un buffer di (almeno)
 * un byte.
 */
ULONG I2CBus::StartRead(UBYTE slave, UBYTE *data, ULONG length)
{
	int temp;
	ULONG len = length;

	if (len > 0)
	{
		// send Start
		if ( (temp = SendStart()) )
		{
			errno = temp;
			return 0;
		}

		if ( (temp = SendByteMast(slave | 1)) != 0 )
		{
			errno = (temp == IICERR_NOTACK) ? IICERR_NOADDRACK : temp;
			last_addr = slave | 1;
			return 0;
		}

		while (len > 1)
		{
			if ((temp = RecByteMast(0)) < 0)
			{
				errno = temp;
				goto fineR;
			}
			*data++ = (UBYTE)temp;
			len--;
		}
	}
	// last byte received without acknowledge
	if ((temp = RecByteMast(1)) < 0)
	{
		errno = temp;
		goto fineR;
	}
	len--;

	*data = (UBYTE)temp;
	errno = 0;

fineR:
	return length-len;
}

ULONG I2CBus::StartWrite(UBYTE slave, UBYTE const *data, ULONG length)
{
	int error;
	ULONG len = length;

	if (len == 0)
		return 0;

	if ( (error = SendStart()) )
	{
		errno = error;
		return 0;
	}
	if ( (error = SendByteMast(slave & 0xFE)) )
	{
		errno = (error == IICERR_NOTACK) ? IICERR_NOADDRACK : error;
		last_addr = slave & 0xFE;
		return 0;
	}

	while (len > 0)
	{
		if ( (error = SendByteMast(*data++)) != 0 )
		{
			errno = error;
			goto fineW;
		}
		len--;
	}

fineW:
	return length-len;
}

int I2CBus::Stop(void)
{
	errno = SendStop() ? IICERR_STOP : 0;

	return errno;
}

int I2CBus::Reset(void)
{
	BYTE c;
	Read(0x00, &c, 0);
	setSCLSDA();
	WaitMsec(100);
	return OK;
}

void I2CBus::Open(int port)
{
	rs232.OpenSerial(port);
	setSCLSDA();
	rs232.SetSerialBreak(1);
	WaitMsec(100);

	Reset();
}

void I2CBus::Close(void)
{
	setSCLSDA();
	rs232.SetSerialBreak(0);
	rs232.CloseSerial();
}

void I2CBus::setSCLSDA()
{
	rs232.SetSerialRTSDTR(1);
}

int I2CBus::isSCLSDAuno() const
{
	return getSDA();
}

void I2CBus::setSDA()
{
	rs232.SetSerialDTR(1);
}

void I2CBus::clearSDA()
{
	rs232.SetSerialDTR(0);
}

void I2CBus::setSCL()
{
	rs232.SetSerialRTS(1);
}

void I2CBus::clearSCL()
{
	rs232.SetSerialRTS(0);
}

int I2CBus::getSDA() const
{
	return rs232.GetSerialCTS();
}

int I2CBus::getSCL() const
{
	return 1;
}

void I2CBus::bitSDA(int bit)
{
	rs232.SetSerialDTR(bit);
}

